from typing import TypedDict, List
from .tvp import TVP

class SeriesPronoDict(TypedDict):
    """
    Parameters:
    -----------
    series_id : int
    
    series_table : str
    
    pronosticos: List[TVP]
    """
    series_id : int
    series_table : str
    pronosticos: List[TVP]



from pandas import DataFrame

class DataFrameDescriptor:
    """DataFrame descriptor with default None"""
    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, instance, owner) -> DataFrame:
        return instance.__dict__[self._name]

    def __set__(self, instance, value):
        try:
            instance.__dict__[self._name] = value if isinstance(value,DataFrame) else DataFrame(value) if value is not None else None
        except ValueError:
            raise ValueError(f'"{self._name}" must be a DataFrame or a DataFrame-coercible type') from None



from ..util import interval2timedelta
from .duration_descriptor import DurationDescriptor
default = None

class DurationDescriptorDefaultNone(DurationDescriptor):
    """Duration descriptor
    Parses dict of unit: value pairs (i.e., {"hours":1,"minutes":30})

    Return type: datetime.timedelta. 
    
    Default: None(hours=0)"""
    def __set__(self, instance, value):
        try:
            instance.__dict__[self._name] = interval2timedelta(value) if value is not None else default
        except ValueError:
            raise ValueError(f'"{self._name}" must be a duration') from None


class ListOrDictDescriptor:
    """List or dict descriptor with default None"""
    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, instance, owner):
        return instance.__dict__[self._name]

    def __set__(self, instance, value):
        if isinstance(value, (list,tuple)):
            try:
                instance.__dict__[self._name] = value if isinstance(value,list) else list(value)
            except ValueError:
                raise ValueError(f'"{self._name}" must be a list or a dict') from None
        elif value is not None:
            try:
                instance.__dict__[self._name] = dict(value)
            except ValueError:
                raise ValueError(f'"{self._name}" must be a list or a dict') from None
        else:
            instance.__dict__[self._name] = None



import collections
from typing import Optional

class TypedList(collections.abc.MutableSequence):

    def __init__(self, oktype, *args, unique_id_property : Optional[str] = None, **fixed_kwargs):
        self.oktype = oktype
        self.list = list()
        self._unique_id_property = unique_id_property
        self._fixed_kwargs = fixed_kwargs
        self.extend(list(args))

    def check(self, v):
        if not isinstance(v, self.oktype):
            raise TypeError("%s must be of type %s" % (str(v),str(self.oktype)))

    def check_else_init(self,v):
        value = None
        if isinstance(v, self.oktype):
            value = v
        else:
            try:
                if isinstance(v,dict):
                    value = self.oktype(**v,**self._fixed_kwargs)
                elif isinstance(v,(list,tuple)):
                    value = self.oktype(*v,**self._fixed_kwargs)
                else:
                    value = self.oktype(v,**self._fixed_kwargs)
            except TypeError as e:
                raise TypeError("Invalid item for mutable sequence. Unable to coerce to type %s" % str(self.oktype),e)
        if self._unique_id_property is not None and getattr(value, self._unique_id_property) in [getattr(o,self._unique_id_property) for o in self.list]:
            raise ValueError("Unique id property %s = %s already present in list" % (self._unique_id_property, getattr(value, self._unique_id_property)))
        return value


    def __len__(self): return len(self.list)

    def __getitem__(self, i): return self.list[i]

    def __delitem__(self, i): del self.list[i]

    def __setitem__(self, i, v):
        v = self.check_else_init(v)
        self.list[i] = v

    def insert(self, i, v):
        v = self.check_else_init(v)
        self.list.insert(i, v)

    def __str__(self):
        return str(self.list)

    def __repr__(self):
        return self.list.__repr__()


from ..util import tryParseAndLocalizeDate

class DatetimeDescriptor:
    """Datetime descriptor
    
    Reads: for absolute date: ISO-8601 datetime string or datetime.datetime. For relative date: dict (duration key-values) or float (decimal number of days). Defaults to None
    
    Returns: None or datetime.datetime"""
    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, instance, owner):
        return instance.__dict__[self._name]

    def __set__(self, instance, value):
        try:
            instance.__dict__[self._name] = tryParseAndLocalizeDate(value) if value is not None else None
        except ValueError:
            raise ValueError(f'"{self._name}" must be a date or a duration') from None



from datetime import datetime, timedelta, date as datetime_date
from typing import Union, Optional
from dateutil.parser import isoparse
from dateutil.relativedelta import relativedelta
from dateutil import tz
import pytz
import logging
from pandas import DatetimeIndex, date_range, DateOffset

def tryParseAndLocalizeDate(
        date_string : Union[str,float,datetime,tuple,datetime_date],
        timezone : str='America/Argentina/Buenos_Aires'
    ) -> Union[datetime, None]:
    """
    Datetime parser. If duration is provided, computes date relative to now.

    Parameters:
    -----------
    date_string : str or float or datetime.datetime
        For absolute date: ISO-8601 datetime string or datetime.datetime.
        For relative date: dict (duration key-values) or float (decimal number of days)
    
    timezone : str
        Time zone string identifier. Default: America/Argentina/Buenos_Aires
    
    Returns:
    --------
    datetime.datetime

    Examples:
    ---------
    ``` 
    tryParseAndLocalizeDate("2024-01-01T03:00:00.000Z")
    tryParseAndLocalizeDate(1.5)
    tryParseAndLocalizeDate({"days":1, "hours": 12}, timezone = "Africa/Casablanca")
    ```
    """
    
    date = isoparse(date_string) if isinstance(date_string,str) else date_string
    if isinstance(date,dict):
        date = datetime.now() + relativedelta(**date)
    elif isinstance(date,(int,float)):
        date = datetime.now() + relativedelta(days=int(date))
    elif isinstance(date, tuple):
        if len(date) < 3:
            raise ValueError("Invalid date tuple: missing items (3 required)")
        date = datetime(*date)
    elif isinstance(date, datetime_date):
        date = datetime(date.year, date.month, date.day)
    if date.tzinfo is None or date.tzinfo.utcoffset(date) is None:
        try:
            date = pytz.timezone(timezone).localize(date)
        except pytz.exceptions.NonExistentTimeError:
            logging.warning("NonexistentTimeError: %s" % str(date))
            return None
    else:
        date = date.astimezone(pytz.timezone(timezone))
    return date

def createDatetimeSequence(
    datetime_index : Optional[DatetimeIndex]=None, 
    timeInterval : Union[relativedelta,dict,int,timedelta] = relativedelta(days=1), 
    timestart : Union[datetime,tuple,str,None] = None, 
    timeend : Union[datetime,tuple,str,None] = None, 
    timeOffset : Union[relativedelta,dict,None] = None
    ) -> DatetimeIndex:
    #Fechas desde timestart a timeend con un paso de timeInterval
    #data: dataframe con index tipo datetime64[ns, America/Argentina/Buenos_Aires]
    #timeOffset sólo para timeInterval n days
    if timestart is None:
        if datetime_index is None:
            raise Exception("Missing datetime_index or timestart+timeend")
        timestart = datetime_index.min()
    else:
        timestart = tryParseAndLocalizeDate(timestart)
    if timestart is None:
        raise pytz.exceptions.NonExistentTimeError("Nonexistent timestart")
    if timeend is None:
        if datetime_index is None:
            raise Exception("Missing datetime_index or timestart+timeend")
        timeend = datetime_index.max()
    else:
        timeend = tryParseAndLocalizeDate(timeend)
    if timeend is None:
        raise pytz.exceptions.NonExistentTimeError("Nonexistent timeend")
    timeInterval = relativedelta(**timeInterval) if isinstance(timeInterval,dict) else timedelta_to_relativedelta(timeInterval) if isinstance(timeInterval,timedelta) else relativedelta(days=timeInterval) if isinstance(timeInterval, int) else timeInterval
    timeOffset = relativedelta(**timeOffset) if isinstance(timeOffset,dict) else timeOffset
    timestart = roundDate(timestart,timeInterval,timeOffset,"up")
    timeend = roundDate(timeend,timeInterval,timeOffset,"down")
    timezone = pytz.timezone("America/Argentina/Buenos_Aires")
    is_subdaily = timeInterval.hours > 0 or timeInterval.minutes > 0 or timeInterval.seconds > 0 or timeInterval.microseconds > 0
    if not is_subdaily:
        if timestart.day == 1 and timeInterval.months > 0:
            freq = "%iMS" % timeInterval.months
        elif timestart.day == 1 and timestart.month == 1 and timeInterval.years > 0:
            freq = "%iYS" % timeInterval.years 
        elif timeInterval.days > 0:
            freq = "%iD" % timeInterval.days
        else:
            freq = relativedelta_to_freq(timeInterval)
        dts_utc = date_range(
            start=timestart.astimezone(tz.UTC), 
            end=timeend.astimezone(tz.UTC), 
            freq = freq
        )
        if timeOffset is not None:
            return DatetimeIndex([timezone.localize(datetime(dt.year,dt.month,dt.day,timeOffset.hours,timeOffset.minutes,timeOffset.seconds)) for dt in dts_utc])
        else:
            return DatetimeIndex([timezone.localize(datetime(dt.year,dt.month,dt.day)) for dt in dts_utc])
    else:
        freq = DateOffset(
                years=timeInterval.years,
                months=timeInterval.months,
                weeks=timeInterval.weeks,
                days=timeInterval.days, 
                hours=timeInterval.hours, 
                minutes = timeInterval.minutes,
                seconds = timeInterval.seconds,
                microseconds = timeInterval.microseconds
            )
        return date_range(
            start=timestart, 
            end=timeend, 
            freq=freq
        ) # .tz_convert(timestart.tzinfo)

def roundDate(date : datetime,timeInterval : relativedelta,timeOffset : Optional[relativedelta]=None, to="up") -> datetime:
    date_0 = tryParseAndLocalizeDate(datetime.combine(date.date(),datetime.min.time()))
    if date_0 is None:
        raise pytz.exceptions.NonExistentTimeError("Nonexistent time")
    if timeOffset is not None:
        date_0 = date_0 + timeOffset 
    while date_0 < date:
        date_0 = date_0 + timeInterval
    if date_0 == date:
        return date_0
    elif to == "up":
        return date_0
    else:
        return date_0 - timeInterval

def interval2timedelta(interval : Union[dict,float,timedelta]):
    """Parses duration dict or number of days into datetime.timedelta object
    
    Parameters:
    -----------
    interval : dict or float (decimal number of days) or datetime.timedelta
        If dict, allowed keys are:
        - days
        - seconds
        - microseconds
        - minutes
        - hours
        - weeks
    
    Returns:
    --------
    duration : datetime.timedelta

    Examples:

    ```
    interval2timedelta({"hours":1, "minutes": 30})
    interval2timedelta(1.5/24)
    ```
    """
    if isinstance(interval,timedelta):
        return interval
    if isinstance(interval,(float,int)):
        return timedelta(days=interval)
    days = 0
    seconds = 0
    microseconds = 0
    milliseconds = 0
    minutes = 0
    hours = 0
    weeks = 0
    for k in interval:
        if k == "milliseconds" or k == "millisecond":
            milliseconds = interval[k]
        elif k == "seconds" or k == "second":
            seconds = interval[k]
        elif k == "minutes" or k == "minute":
            minutes = interval[k]
        elif k == "hours" or k == "hour":
            hours = interval[k]
        elif k == "days" or k == "day":
            days = interval[k]
        elif k == "weeks" or k == "week":
            weeks = interval[k] * 86400 * 7
    return timedelta(days=days, seconds=seconds, microseconds=microseconds, milliseconds=milliseconds, minutes=minutes, hours=hours, weeks=weeks)

def timedelta_to_relativedelta(td: timedelta) -> relativedelta:
    total_seconds = td.total_seconds()
    
    days, remainder = divmod(total_seconds, 86400)  # seconds in a day
    hours, remainder = divmod(remainder, 3600)
    minutes, seconds = divmod(remainder, 60)

    return relativedelta(
        days=int(days),
        hours=int(hours),
        minutes=int(minutes),
        seconds=int(seconds)
    )

def relativedelta_to_timedelta(rd: relativedelta) -> timedelta:
    if not isinstance(rd, relativedelta):
        raise TypeError("Value must be of type relativedelta")
    total_seconds = relativedeltaToSeconds(rd)
    days, remainder = divmod(total_seconds, 86400)  # seconds in a day
    hours, remainder = divmod(remainder, 3600)
    minutes, seconds = divmod(remainder, 60)

    return timedelta(
        days=int(days),
        hours=int(hours),
        minutes=int(minutes),
        seconds=int(seconds)
    )

def relativedeltaToSeconds(rd : relativedelta) -> int:
    if not isinstance(rd, relativedelta):
        raise TypeError("Value must be of type relativedelta")
    now = datetime.now()
    future = now + rd
    delta = future - now
    return int(delta.total_seconds())

def relativedelta_to_freq(rd: relativedelta) -> str:
    if rd.years:
        return f"{rd.years}Y"
    if rd.months:
        return f"{rd.months}M"
    if rd.days:
        return f"{rd.days}D"
    if rd.hours:
        return f"{rd.hours}H"
    if rd.minutes:
        return f"{rd.minutes}T"
    if rd.seconds:
        return f"{rd.seconds}S"
    raise ValueError("Unsupported relativedelta")



from .bool_descriptor import BoolDescriptor
from .bool_or_none_descriptor import BoolOrNoneDescriptor
from .dataframe_descriptor import DataFrameDescriptor
from .datetime_descriptor import DatetimeDescriptor
from .dict_descriptor import DictDescriptor
from .duration_descriptor import DurationDescriptor
from .filepath_descriptor import FilepathDescriptor
from .float_descriptor import FloatDescriptor
from .int_descriptor import IntDescriptor
from .list_descriptor import ListDescriptor
from .list_or_dict_descriptor import ListOrDictDescriptor
from .string_descriptor import StringDescriptor

__all__ = ['BoolDescriptor','BoolOrNoneDescriptor','DataFrameDescriptor','DatetimeDescriptor','DictDescriptor','DurationDescriptor','FilepathDescriptor','FloatDescriptor','IntDescriptor','ListDescriptor','ListOrDictDescriptor','StringDescriptor']


from datetime import timedelta
from ..util import interval2timedelta

default = timedelta(hours=0)

class DurationDescriptor:
    """Duration descriptor
    Parses dict of unit: value pairs (i.e., {"hours":1,"minutes":30})

    Return type: datetime.timedelta. 
    
    Default: timedelta(hours=0)"""
    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, instance, owner):
        return instance.__dict__[self._name]

    def __set__(self, instance, value):
        try:
            instance.__dict__[self._name] = interval2timedelta(value) if value is not None else default
        except ValueError:
            raise ValueError(f'"{self._name}" must be a duration') from None


from typing import TypedDict, List
from .series_prono_dict import SeriesPronoDict
from datetime import datetime

class CorridaDict(TypedDict):
    """
    Parameters:
    -----------
    cal_id : int
    
    id : int

    forecast_date : datetime
    
    series: List[SeriesPronoDict]
    """
    cal_id : int
    id : int
    forecast_date : datetime
    series: List[SeriesPronoDict]



from typing import TypedDict
from datetime import datetime

class TVP(TypedDict):
    """
    Parameters:
    -----------
    timestart : datetime
    
    valor : float

    series_id : int = None
    """
    timestart : datetime
    valor : float
    series_id : int



class BoolDescriptor:
    """Boolean attribute default False"""
    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, instance, owner):
        return instance.__dict__[self._name]

    def __set__(self, instance, value):
        try:
            instance.__dict__[self._name] = bool(value)
        except ValueError:
            raise ValueError(f'"{self._name}" must be a boolean') from None



class StringDescriptor:
    def __set_name__(self, owner, name):
        """A string descriptor with default None"""
        self._name = name

    def __get__(self, instance, owner):
        return instance.__dict__[self._name]

    def __set__(self, instance, value):
        try:
            instance.__dict__[self._name] = str(value) if value is not None else None
        except ValueError:
            raise ValueError(f'"{self._name}" must be a string') from None



from .api_config_dict import ApiConfigDict
from .corrida_dict import CorridaDict
from .series_dict import SeriesDict
from .series_prono_dict import SeriesPronoDict
from .tvp import TVP
from .typed_list import TypedList

__all__ = ['ApiConfigDict','CorridaDict','SeriesDict','TVP','SeriesPronoDict','TypedList']


from setuptools import setup, find_packages

setup(
    name='a5_client_utils',
    version='0.1.2',
    packages=find_packages(),
    description='Auxiliary functions for a5 API clients',
    author='Juan F. Bianchi',
    author_email='jbianchi@ina.gob.ar',
    url='https://github.com/jbianchi81/a5-client-utils',  # Optional
)



from typing import TypedDict

class ApiConfigDict(TypedDict):
    """
        url : str
            api base url
        token : str    
            api authorization token
    """
    url : str
    token : str


class BoolOrNoneDescriptor:
    """Boolean or None attribute default None"""
    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, instance, owner):
        return instance.__dict__[self._name]

    def __set__(self, instance, value):
        if value is None:
            instance.__dict__[self._name] = None
        else:
            try:
                instance.__dict__[self._name] = bool(value)
            except ValueError:
                raise ValueError(f'"{self._name}" must be a boolean or None') from None
